import json
from pathlib import Path

from ruamel.yaml import YAML

from .capacities import extract_capacities
from .cluster import get_cluster as _get_cluster
from .monitoring import extract_monitoring as _extract_monitoring
from .policies import get_qos as _get_qos
from .policies import get_reconfiguration as _get_reconfiguration
from .policies import get_scheduling as _get_scheduling
from .rdt import generate_rdt as _generate_rdt
from .requirements import tosca_to_ask_dict
from .validation import (
    TemplateKind,
    classify_template,
    requires_kind,
    validate_template,
)

yaml = YAML(typ="safe")


class DotDict:
    def __init__(self, **entries):
        for k, v in entries.items():
            if isinstance(v, dict):
                v = DotDict(**v)
            elif isinstance(v, list):
                v = [DotDict(**i) if isinstance(i, dict) else i for i in v]
            setattr(self, k, v)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __delitem__(self, key):
        delattr(self, key)

    def __contains__(self, key):
        return hasattr(self, key)

    def __repr__(self):
        return repr(self._to_dict())

    def _to_dict(self):
        result = {}
        for key, value in self.__dict__.items():
            if isinstance(value, DotDict):
                result[key] = value._to_dict()
            elif isinstance(value, list):
                result[key] = [
                    v._to_dict() if isinstance(v, DotDict) else v for v in value
                ]
            else:
                result[key] = value
        return result

    def _to_json(self, indent=None, **kwargs):
        return json.dumps(self._to_dict(), indent=indent, **kwargs)


class Sardou(DotDict):
    def __init__(self, path=None, content=None):
        if path is None and content is None:
            raise ValueError("Either 'path' or 'content' must be provided")

        if path is not None and content is not None:
            raise ValueError("Cannot provide both 'path' and 'content'")

        if path is not None:
            # File-based initialization
            path = Path(path)
            if not path.exists():
                raise FileNotFoundError(f"File does not exist: {path}")
            self.path = path
            source = path
        else:
            self.path = None
            source = content

        template = validate_template(source)
        if not template:
            label = str(path) if path else "provided content"
            raise ValueError(f"Validation failed for: {label}")

        resolved = yaml.load(template.stdout)
        super().__init__(**resolved)

        if path is not None:
            with path.open("r") as f:
                raw = yaml.load(f)
        elif isinstance(content, dict):
            raw = content
        else:
            raw = yaml.load(content)

        self.kind = classify_template(self)
        self.raw = DotDict(**raw)

    def get_requirements(self):
        return tosca_to_ask_dict(self.raw._to_dict())

    @requires_kind(TemplateKind.SAT)
    def get_qos(self):
        return _get_qos(self)

    @requires_kind(TemplateKind.SAT)
    def get_reconfiguration(self):
        return _get_reconfiguration(self)

    @requires_kind(TemplateKind.SAT)
    def get_scheduling(self):
        return _get_scheduling(self)

    @requires_kind(TemplateKind.CDT)
    def get_capacities(self):
        nodes = self.nodeTemplates._to_dict()
        return extract_capacities(nodes)

    def generate_rdt(self, selected_offer, output_path="rdt.yaml"):
        return _generate_rdt(self, selected_offer, output_path=output_path)

    @requires_kind(TemplateKind.RDT)
    def get_cluster(self, resource_suffix=None):
        return _get_cluster(self, resource_suffix=resource_suffix)

    @requires_kind(TemplateKind.SAT)
    def get_monitoring(self):
        nodes = self.raw.service_template.node_templates
        return _extract_monitoring(nodes._to_dict())




"""
TOSCA Resource Requirements Extractor
Converts TOSCA node_filter data to expression + colocation format
"""

OPERATOR_MAP = {
    "$greater_or_equal": ">=",
    "$greater_than": ">",
    "$less_or_equal": "<=",
    "$less_than": "<",
    "$equal": "==",
    "$not_equal": "!=",
    "$has_any_entry": "has_any_entry",
}

COLOCATION_POLICY = "Scheduling.Colocation"


def get_properties(relationship):
    """Convert properties in relationship to node properties"""
    properties = relationship.get("properties", {})
    if not isinstance(properties, dict):
        return {}

    return properties


def build_expression(node_filter):
    """Convert TOSCA node_filter to a lambda expression string."""
    conditions = node_filter.get("$and", [])
    parts = []

    for condition in conditions:
        for constraint_func, args in condition.items():
            if len(args) < 2:
                continue
            property_path = args[0].get("$get_property", [])
            if len(property_path) < 5:
                continue

            capability_type = property_path[3]
            property_name = property_path[4]
            value = args[1]
            op = OPERATOR_MAP.get(constraint_func, "==")

            key = f"{capability_type}.{property_name}"
            if op == "has_any_entry":
                parts.append(f"(any(entry in vals['{key}'] for entry in {value}))")
            elif isinstance(value, str):
                parts.append(f"(vals['{key}'] {op} '{value}')")
            else:
                parts.append(f"(vals['{key}'] {op} {value})")

    body = " and ".join(parts)
    return f"lambda vals: ({body})"


def extract_colocation_groups(tosca_dict):
    """Extract colocation groups from policies."""
    groups = []
    policies = tosca_dict.get("service_template", {}).get("policies", [])
    for policy in policies:
        for policy_name, policy_data in policy.items():
            if policy_data.get("type", "").endswith(COLOCATION_POLICY):
                targets = policy_data.get("targets", [])
                if targets:
                    groups.append(targets)
    return groups


def extract_reqs_with_filter(tosca_dict):
    reqs_with_filter = {}
    node_templates = tosca_dict.get("service_template", {}).get("node_templates", {})
    for node_name, node_data in node_templates.items():
        for req in node_data.get("requirements", []):
            for _, req_data in req.items():
                if "node_filter" in req_data:
                    reqs_with_filter[node_name] = req_data
    return reqs_with_filter


def tosca_to_ask_dict(tosca_dict):
    """
    Convert TOSCA dict to ask format dict with expressions and colocation.
    """
    reqs_with_filter = extract_reqs_with_filter(tosca_dict)
    colocation_groups = extract_colocation_groups(tosca_dict)

    # Build a map: node -> (representative, list of colocated peers)
    node_to_representative = {}
    representative_colocated = {}

    for group in colocation_groups:
        rep = group[0]
        others = group[1:]
        representative_colocated[rep] = others
        for node in group:
            node_to_representative[node] = rep

    result = {}
    for node_name, req_data in reqs_with_filter.items():
        # Skip non-representative nodes in a colocation group
        if (
            node_name in node_to_representative
            and node_to_representative[node_name] != node_name
        ):
            continue

        node_filter = req_data["node_filter"]
        relationship = req_data.get("relationship", {})
        result[node_name] = {
            "expression": build_expression(node_filter),
            "colocated": representative_colocated.get(node_name, []),
            "properties": get_properties(relationship),
        }

    return result



def flatten(val):
    if isinstance(val, dict):
        if "$primitive" in val:
            return val["$primitive"]
        if "$list" in val:
            return [flatten(v) for v in val["$list"]]
        if "$map" in val:
            out = {}
            for pair in val["$map"]:
                key = flatten(pair.get("$key"))
                value = flatten(pair)
                out[key] = value
            return out
        return {k: flatten(v) for k, v in val.items()}
    return val


def extract_properties(prop_dict):
    extracted = {}
    for k, v in prop_dict.items():
        meta = v.get("$meta", {}) if isinstance(v, dict) else {}
        alias = meta.get("metadata", {}).get("alias")
        extracted_key = alias if alias else k
        extracted[extracted_key] = flatten(v)
    return extracted



import os

from .utils import extract_properties

LABELS_PREFIX = "labels.swarmchestrate.eu/"

def get_cluster(rdt, resource_suffix=None):
    resource_suffix = (
        resource_suffix or os.getenv("TOSCA_RESOURCE_SUFFIX") or "::Capacity"
    )

    resources = {}

    for name, node in rdt.nodeTemplates._to_dict().items():
        is_resource = any(
            t.get("parent", "").endswith(resource_suffix) or k.endswith(resource_suffix)
            for k, t in node.get("types", {}).items()
        )

        if not is_resource:
            continue

        extracted = extract_properties(node.get("properties", {}))

        extracted["node_labels"] = {
            f"{LABELS_PREFIX}ms_id": node.get("metadata", {}).get("ms_id")
        }
        resources[name] = extracted

    return resources



from importlib.metadata import PackageNotFoundError, version

from .sardou import Sardou  # noqa: F401

try:
    __version__ = version("sardou")
except PackageNotFoundError:
    __version__ = "unknown"



_MONITORING_KEYS = ("metrics", "slo-constraints")


def _extract_monitoring(capabilities: dict) -> dict:
    """Extract monitoring properties from a single node's capabilities.

    If either key is present, the other is included as an empty dict.
    """
    monitoring = {
        key: capabilities.get(key, {}).get("properties", {}) for key in _MONITORING_KEYS
    }
    if not any(monitoring.values()):
        return {}
    return monitoring


def extract_monitoring(node_templates: dict) -> dict:
    """Return monitoring info (metrics / slo-constraints) per node.

    Nodes that define neither capability are omitted from the result.
    """
    result = {}
    for name, node in node_templates.items():
        capabilities = node.get("capabilities", {}) or {}
        monitoring = _extract_monitoring(capabilities)
        if monitoring:
            result[name] = monitoring
    return result



from pathlib import Path

import jinja2
from ruamel.yaml import YAML

yaml = YAML(typ="safe")

jinja_loader = jinja2.FileSystemLoader(searchpath="tools/jinja_templates")
jinja_env = jinja2.Environment(loader=jinja_loader)
template = jinja_env.get_template("docspage.md.j2")

DOCS_LOC = Path("docs/")

SOURCES = {
    "app": "profiles/eu.swarmchestrate/app.yaml",
    "capacity": "profiles/eu.swarmchestrate/capacity.yaml",
    "monitoring": "profiles/eu.swarmchestrate/monitoring.yaml",
    "policy": "profiles/eu.swarmchestrate/policy.yaml",
}

# --- Load YAML sources ---

data = {}
for key, path in SOURCES.items():
    with open(path) as f:
        data[key] = yaml.load(f)


# --- Extraction helpers ---


def extract_fields(properties):
    """Convert a YAML properties dict to template `fields` format."""
    fields = {}
    for name, value in properties.items():
        entry_schema = value.get("entry_schema", {})
        fields[name] = {
            "required": value.get("required", False),
            "type": value.get("type", ""),
            "schema": (
                entry_schema
                if isinstance(entry_schema, str)
                else entry_schema.get("type", "")
            ),
            "description": value.get("description", "No description available"),
        }
    return fields


def extract_types(types_dict):
    """Convert a data_types / policy_types dict to template `types` format."""
    types = {}
    for name, value in types_dict.items():
        types[name] = {
            "description": value.get("description", "No description available"),
            "fields": [
                {
                    "name": prop_name,
                    "required": prop_value.get("required", False),
                    "type": prop_value.get("type", ""),
                    "description": prop_value.get(
                        "description", "No description available"
                    ),
                }
                for prop_name, prop_value in value.get("properties", {}).items()
            ],
        }
    return types


def extract_capacity_fields(node_types_dict, capability_types_dict):
    """Convert node_types to fields format with grouped capabilities.

    Each node type becomes a top-level field.  Its capabilities are grouped
    so that e.g. ``locality`` appears as a heading with ``continent``,
    ``city``, etc. nested underneath (instead of flat ``locality.continent``).
    """
    fields = {}
    for node_name, node_value in node_types_dict.items():
        # Node-level properties (e.g. ip, image_id)
        node_props = [
            {
                "name": prop_name,
                "required": prop_value.get("required", False),
                "type": prop_value.get("type", ""),
                "description": prop_value.get(
                    "description", "No description available"
                ),
            }
            for prop_name, prop_value in node_value.get("properties", {}).items()
        ]

        # Capabilities with their properties grouped
        capabilities = []
        for cap_name, cap_value in node_value.get("capabilities", {}).items():
            props = cap_value.get("properties", {})
            if not props:
                cap_type = cap_value.get("type", "")
                if cap_type in capability_types_dict:
                    props = capability_types_dict[cap_type].get("properties", {})

            cap_fields = [
                {
                    "name": prop_name,
                    "required": prop_value.get("required", False),
                    "type": prop_value.get("type", ""),
                    "default": prop_value.get("default", ""),
                    "description": prop_value.get(
                        "description", "No description available"
                    ),
                }
                for prop_name, prop_value in props.items()
            ]
            if cap_fields:
                cap_desc = cap_value.get("description", "")
                capabilities.append(
                    {"name": cap_name, "description": cap_desc, "fields": cap_fields}
                )

        fields[node_name] = {
            "description": node_value.get("description", "No description available"),
            "capabilities": capabilities,
            "props": node_props,
        }
    return fields


# --- Page definitions ---

pages = [
    {
        "name": "Policy",
        "fields": {
            "targets": {
                "required": False,
                "type": "list",
                "entry_schema": "string",
                "description": "**Present on all policies that target specific Microservices**. A list of Microservice names to target with this policy.",
            }
        },
        "types": extract_types(data["policy"].get("policy_types", {})),
    },
    {
        "name": "Microservice",
        "fields": extract_fields(
            data["app"]["node_types"]["Microservice"]["properties"]
        ),
        "types": extract_types(data["app"].get("data_types", {})),
        "show_primitives": True,
    },
    {
        "name": "Capacity",
        "fields": extract_capacity_fields(
            data["capacity"].get("node_types", {}),
            data["capacity"].get("capability_types", {}),
        ),
    },
    {
        "name": "Monitoring",
        "fields": extract_fields(
            data["app"]["node_types"]["Microservice"]["capabilities"]
        ),
        "types": extract_types(data["monitoring"].get("data_types", {})),
    },
]


# --- Render and write ---

for page in pages:
    page.setdefault("fields", {})
    page.setdefault("types", {})
    page.setdefault("show_primitives", False)

    rendered = template.render(**page)
    file_name = page["name"].lower().replace(" ", "_")

    write_path = DOCS_LOC / f"{file_name}.md"
    with open(write_path, "w") as f:
        f.write(rendered)



import argparse
import glob
import logging
import sys
import traceback
from pathlib import Path

from sardou import Sardou


def main():
    logging.basicConfig(level=logging.INFO, format="%(message)s")

    parser = argparse.ArgumentParser(
        description="Sardou - TOSCA template parser and processor"
    )
    parser.add_argument(
        "path",
        type=str,
        nargs="+",
        help="Path(s) to TOSCA template file(s), supports glob patterns",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Enable verbose output"
    )

    args = parser.parse_args()

    # Expand glob patterns and collect all files
    files = []
    for path_pattern in args.path:
        expanded = glob.glob(path_pattern)
        if expanded:
            files.extend(expanded)
        else:
            # If glob returns nothing, treat as literal path
            files.append(path_pattern)

    if not files:
        print("Error: No files found matching the given path(s)", file=sys.stderr)
        sys.exit(1)

    # Process each file
    for file_path in files:
        target_path = Path(file_path).resolve()
        if not target_path.exists():
            print(f"Error: Path does not exist: {file_path}", file=sys.stderr)
            sys.exit(1)

        if args.verbose:
            print(f"Parsing TOSCA template: {target_path}")

        try:
            Sardou(str(target_path))

            if args.verbose:
                print("Successfully parsed TOSCA template")
        except Exception as e:
            if args.verbose:
                traceback.print_exc(file=sys.stderr)
            else:
                print(f"Error parsing TOSCA template: {e}", file=sys.stderr)
            sys.exit(1)

    sys.exit(0)


if __name__ == "__main__":
    main()



"""Unit tests for package internals that don't require puccini-tosca."""

import pytest

# ---------------------------------------------------------------------------
# Import / metadata
# ---------------------------------------------------------------------------


class TestImport:
    def test_sardou_is_importable(self):
        import sardou  # noqa: F401

    def test_version_is_a_string(self):
        import sardou

        assert isinstance(sardou.__version__, str)

    def test_version_not_unknown(self):
        import sardou

        assert sardou.__version__ != "unknown"

    def test_sardou_class_exported(self):
        from sardou import Sardou

        assert callable(Sardou)


# ---------------------------------------------------------------------------
# DotDict
# ---------------------------------------------------------------------------


class TestDotDict:
    @pytest.fixture
    def DotDict(self):
        from sardou.sardou import DotDict

        return DotDict

    def test_attribute_access(self, DotDict):
        d = DotDict(foo="bar")
        assert d.foo == "bar"

    def test_nested_dict_becomes_dotdict(self, DotDict):
        d = DotDict(outer={"inner": 42})
        assert isinstance(d.outer, DotDict)
        assert d.outer.inner == 42

    def test_list_of_dicts_converted(self, DotDict):
        d = DotDict(items=[{"x": 1}, {"x": 2}])
        assert isinstance(d.items[0], DotDict)
        assert d.items[1].x == 2

    def test_contains(self, DotDict):
        d = DotDict(a=1)
        assert "a" in d
        assert "b" not in d

    def test_getitem(self, DotDict):
        d = DotDict(key="value")
        assert d["key"] == "value"

    def test_setitem(self, DotDict):
        d = DotDict(key="old")
        d["key"] = "new"
        assert d.key == "new"

    def test_delitem(self, DotDict):
        d = DotDict(key="value")
        del d["key"]
        assert "key" not in d

    def test_to_dict_roundtrip(self, DotDict):
        original = {"a": 1, "b": {"c": 2}, "d": [{"e": 3}]}
        d = DotDict(**original)
        assert d._to_dict() == original

    def test_to_json(self, DotDict):
        import json

        d = DotDict(x=1)
        parsed = json.loads(d._to_json())
        assert parsed == {"x": 1}


# ---------------------------------------------------------------------------
# prevalidate
# ---------------------------------------------------------------------------


class TestPrevalidate:
    @pytest.fixture
    def prevalidate(self):
        from sardou.validation import prevalidate

        return prevalidate

    def test_nonexistent_file_returns_false(self, prevalidate, tmp_path):
        result = prevalidate(tmp_path / "missing.yaml")
        assert result is False

    def test_empty_file_returns_false(self, prevalidate, tmp_path):
        f = tmp_path / "empty.yaml"
        f.write_text("")
        assert prevalidate(f) is False

    def test_valid_minimal_template_returns_data(self, prevalidate, tmp_path):
        """prevalidate should return the parsed data for a structurally valid file."""
        f = tmp_path / "minimal.yaml"
        f.write_text(
            "tosca_definitions_version: tosca_2_0\n"
            "imports: []\n"
            "service_template:\n"
            "  node_templates: {}\n"
        )
        result = prevalidate(f)
        assert result is not False
        assert "service_template" in result

    def test_profile_key_renamed_to_url(self, prevalidate, tmp_path):
        """prevalidate should rewrite 'profile' import keys to 'url'."""
        f = tmp_path / "profile_import.yaml"
        f.write_text(
            "tosca_definitions_version: tosca_2_0\n"
            "imports:\n"
            "  - profile: some.profile\n"
            "service_template:\n"
            "  node_templates: {}\n"
        )
        result = prevalidate(f)
        assert result is not False
        assert result["imports"][0].get("url") == "some.profile"
        assert "profile" not in result["imports"][0]



"""Shared pytest marks for the test suite."""

import shutil

import pytest

requires_puccini = pytest.mark.skipif(
    shutil.which("puccini-tosca") is None,
    reason="puccini-tosca not found in PATH",
)



_CAPACITY_KEY = "capacity"


def _unwrap(value):
    if isinstance(value, dict):
        if "$primitive" in value:
            return value["$primitive"]
        if "$list" in value:
            return [_unwrap(item) for item in value["$list"]]
        if "$map" in value:
            return dict(_unwrap_map_entry(entry) for entry in value["$map"])
    return value


def _unwrap_map_entry(entry):
    key = _unwrap(entry["$key"])
    value = _unwrap({k: v for k, v in entry.items() if k != "$key"})
    return key, value


def _is_overall(node: dict) -> bool:
    types = node.get("types") or {}
    return any("OverallCapacity" in type_name for type_name in types)


def _get_res_key(node: dict) -> str:
    res_type = (
        _unwrap(
            node.get("capabilities", {})
            .get("resource", {})
            .get("properties", {})
            .get("type", {})
        )
        or ""
    )

    return "edge_instances" if "edge" in res_type.lower() else "cloud_flavours"


def _process_node(name, node, capabilities, capacities, capacity_by):
    res_key = _get_res_key(node)
    capacities.setdefault(res_key, {})[name] = {}

    for cap_name, capability in capabilities.items():
        props = capability.get("properties", {}) or {}
        if props and cap_name != _CAPACITY_KEY:
            capacities[res_key][name][cap_name] = {
                k: _unwrap(v) for k, v in props.items()
            }

    instances = (
        capabilities.get(_CAPACITY_KEY, {}).get("properties", {}).get("instances")
    )
    if res_key == "cloud_flavours":
        capacity_by[name] = _unwrap(instances) if instances is not None else 1


def extract_capacities(processed_nodes: dict) -> dict:
    capacity_by = {}
    capacities = {}
    overall = None

    for name, node in processed_nodes.items():
        capabilities = node.get("capabilities", {}) or {}
        if _is_overall(node):
            cap_props = capabilities.get(_CAPACITY_KEY, {}).get("properties", {})
            overall = {k: _unwrap(v) for k, v in cap_props.items()}
            continue

        _process_node(name, node, capabilities, capacities, capacity_by)

    if overall is not None:
        capacities["cloud_capacity_raw"] = overall
    elif capacity_by:
        capacities["cloud_capacity_flavour"] = capacity_by

    return capacities



import sys
from pathlib import Path

from ruamel.yaml import YAML

from sardou.manifestGenerator import get_kubernetes_manifest

yaml_parser = YAML()
yaml_parser.default_flow_style = False

TOSCA_FILE = "templates/tosca-filename.yaml"
OUTPUT_FILE = "manifest-out.yaml"
IMAGE_PULL_SECRET = "my-registry-secret"  # Optional: set to None if not needed

path = Path(TOSCA_FILE)
if not path.exists():
    sys.exit(f"Error: TOSCA file '{TOSCA_FILE}' not found.")

try:
    manifests = get_kubernetes_manifest(TOSCA_FILE, image_pull_secret=IMAGE_PULL_SECRET)

    if not manifests:
        sys.exit("Warning: No Kubernetes manifests generated.")

    with open(OUTPUT_FILE, "w") as f:
        yaml_parser.dump_all(manifests, f)

    print(
        f"✅ Kubernetes manifests written to '{OUTPUT_FILE}' ({len(manifests)} items)"
    )

except Exception as e:
    sys.exit(f"Error: {e}")



from .utils import extract_properties

_PREFIX = "eu.swarmchestrate"


def _get_policies(sat, suffix, incl_type=False):
    if not hasattr(sat, "policies"):
        return {}

    policies = sat.policies._to_dict()
    result = {}

    for name, policy in policies.items():
        types = policy.get("types", {})
        match = any(k.startswith(_PREFIX) and k.endswith(suffix) for k in types)
        if not match:
            continue

        policy_data = extract_properties(policy.get("properties", {}))

        if incl_type:
            policy_data["type"] = list(types.keys())[-1]

        raw_policies = sat.raw._to_dict()["service_template"]["policies"]
        for raw_policy in raw_policies:
            if name not in raw_policy:
                continue

            targets = raw_policy[name].get("targets")
            if targets:
                policy_data["targets"] = targets

        result[name] = policy_data

    return result


def get_qos(sat):
    return _get_policies(sat, "QoS", incl_type=True)


def get_reconfiguration(sat):
    return _get_policies(sat, "Reconfiguration")


def get_scheduling(sat):
    return _get_policies(sat, "Scheduling", incl_type=True)



"""CLI smoke tests"""

import subprocess
from pathlib import Path

from tests.marks import requires_puccini

CDT_DIR = Path(__file__).parent / "templates" / "cdt"
SAT_DIR = Path(__file__).parent / "templates" / "sat"


def run(*args, **kwargs):
    """Run the sardou CLI and return the CompletedProcess."""
    return subprocess.run(
        ["sardou", *args],
        capture_output=True,
        text=True,
        **kwargs,
    )


class TestHelp:
    def test_help_exits_zero(self):
        result = run("--help")
        assert result.returncode == 0


class TestBadArguments:
    def test_no_args_exits_nonzero(self):
        result = run()
        assert result.returncode != 0

    def test_missing_file_exits_one(self, tmp_path):
        result = run(str(tmp_path / "not_there.yaml"))
        assert result.returncode == 1

    def test_missing_file_prints_error(self, tmp_path):
        result = run(str(tmp_path / "not_there.yaml"))
        assert "exist" in result.stderr.lower() and "error" in result.stderr.lower()


class TestInvalidYaml:
    def test_empty_file_exits_one(self, tmp_path):
        f = tmp_path / "empty.yaml"
        f.write_text("")
        result = run(str(f))
        assert result.returncode == 1

    def test_non_yaml_file_exits_one(self, tmp_path):
        f = tmp_path / "junk.yaml"
        f.write_text(":::not valid yaml:::")
        result = run(str(f))
        assert result.returncode == 1


@requires_puccini
class TestCDTTemplates:
    """Smoke-test that each CDT template passes sardou parsing end-to-end."""

    def test_cloud_overall_parses(self):
        result = run(str(CDT_DIR / "cloud-overall.yaml"))
        assert result.returncode == 0, result.stderr

    def test_cloud_instances_parses(self):
        result = run(str(CDT_DIR / "cloud-instances.yaml"))
        assert result.returncode == 0, result.stderr

    def test_edge_parses(self):
        result = run(str(CDT_DIR / "edge.yaml"))
        assert result.returncode == 0, result.stderr


@requires_puccini
class TestSATTemplates:
    """Smoke-test that each SAT template passes sardou parsing end-to-end."""

    def test_bookinfo_parses(self):
        result = run(str(SAT_DIR / "BookInfo.yaml"))
        assert result.returncode == 0, result.stderr



"""Tests for the sardou library API"""

from pathlib import Path

import pytest
from ruamel.yaml import YAML

from sardou import Sardou
from sardou.sardou import Sardou as SardouInternal
from sardou.validation import TemplateKind
from tests.marks import requires_puccini

CDT_DIR = Path(__file__).parent / "templates" / "cdt"
SAT_DIR = Path(__file__).parent / "templates" / "sat"

_yaml = YAML(typ="safe")


def _load_sardou(filepath, mode):
    """Create a Sardou instance from *filepath* using the given input mode."""
    if mode == "path":
        return Sardou(str(filepath))
    # mode == "dict"
    with open(filepath) as f:
        data = _yaml.load(f)
    return Sardou(content=data)


# ---------------------------------------------------------------------------
# requirements.py — build_expression
# ---------------------------------------------------------------------------


class TestBuildExpression:
    @pytest.fixture
    def build(self):
        from sardou.requirements import build_expression

        return build_expression

    def test_empty_filter_returns_empty_lambda(self, build):
        assert build({}) == "lambda vals: ()"

    def test_no_and_key_returns_empty_lambda(self, build):
        assert build({"$or": []}) == "lambda vals: ()"

    def test_equal_constraint(self, build):
        node_filter = {
            "$and": [
                {
                    "$equal": [
                        {
                            "$get_property": [
                                "SELF",
                                "TARGET",
                                "CAPABILITY",
                                "host",
                                "num-cpus",
                            ]
                        },
                        4,
                    ]
                }
            ]
        }
        result = build(node_filter)
        assert result == "lambda vals: ((vals['host.num-cpus'] == 4))"

    def test_greater_or_equal_constraint(self, build):
        node_filter = {
            "$and": [
                {
                    "$greater_or_equal": [
                        {
                            "$get_property": [
                                "SELF",
                                "TARGET",
                                "CAPABILITY",
                                "host",
                                "mem-size",
                            ]
                        },
                        2,
                    ]
                }
            ]
        }
        result = build(node_filter)
        assert result == "lambda vals: ((vals['host.mem-size'] >= 2))"

    def test_string_value_quoted(self, build):
        node_filter = {
            "$and": [
                {
                    "$equal": [
                        {
                            "$get_property": [
                                "SELF",
                                "TARGET",
                                "CAPABILITY",
                                "locality",
                                "city",
                            ]
                        },
                        "budapest",
                    ]
                }
            ]
        }
        result = build(node_filter)
        assert result == "lambda vals: ((vals['locality.city'] == 'budapest'))"

    def test_multiple_conditions_joined_with_and(self, build):
        node_filter = {
            "$and": [
                {
                    "$greater_or_equal": [
                        {
                            "$get_property": [
                                "SELF",
                                "TARGET",
                                "CAPABILITY",
                                "host",
                                "num-cpus",
                            ]
                        },
                        2,
                    ]
                },
                {
                    "$greater_or_equal": [
                        {
                            "$get_property": [
                                "SELF",
                                "TARGET",
                                "CAPABILITY",
                                "host",
                                "mem-size",
                            ]
                        },
                        4,
                    ]
                },
            ]
        }
        result = build(node_filter)
        assert (
            result
            == "lambda vals: ((vals['host.num-cpus'] >= 2) and (vals['host.mem-size'] >= 4))"
        )

    def test_short_property_path_skipped(self, build):
        node_filter = {
            "$and": [
                {
                    "$equal": [
                        {"$get_property": ["SELF", "TARGET"]},
                        4,
                    ]
                }
            ]
        }
        assert build(node_filter) == "lambda vals: ()"


# ---------------------------------------------------------------------------
# requirements.py — extract_colocation_groups
# ---------------------------------------------------------------------------


class TestExtractColocationGroups:
    @pytest.fixture
    def extract(self):
        from sardou.requirements import extract_colocation_groups

        return extract_colocation_groups

    def test_no_policies_returns_empty(self, extract):
        assert extract({"service_template": {}}) == []

    def test_non_colocation_policy_ignored(self, extract):
        tosca = {
            "service_template": {
                "policies": [
                    {
                        "p1": {
                            "type": "swch:Scheduling.AntiColocation",
                            "targets": ["a", "b"],
                        }
                    }
                ]
            }
        }
        assert extract(tosca) == []

    def test_colocation_targets_extracted(self, extract):
        tosca = {
            "service_template": {
                "policies": [
                    {
                        "p1": {
                            "type": "swch:Scheduling.Colocation",
                            "targets": ["a", "b", "c"],
                        }
                    }
                ]
            }
        }
        result = extract(tosca)
        assert result == [["a", "b", "c"]]

    def test_multiple_colocation_groups(self, extract):
        tosca = {
            "service_template": {
                "policies": [
                    {
                        "p1": {
                            "type": "swch:Scheduling.Colocation",
                            "targets": ["a", "b"],
                        }
                    },
                    {
                        "p2": {
                            "type": "swch:Scheduling.Colocation",
                            "targets": ["c", "d"],
                        }
                    },
                ]
            }
        }
        result = extract(tosca)
        assert len(result) == 2

    def test_empty_dict_returns_empty(self, extract):
        assert extract({}) == []


# ---------------------------------------------------------------------------
# requirements.py — extract_reqs_with_filter
# ---------------------------------------------------------------------------


class TestExtractNodesWithFilter:
    @pytest.fixture
    def extract(self):
        from sardou.requirements import extract_reqs_with_filter

        return extract_reqs_with_filter

    def _tosca(self, node_templates):
        return {"service_template": {"node_templates": node_templates}}

    def test_no_requirements_returns_empty(self, extract):
        tosca = self._tosca({"app": {"type": "swch:Microservice"}})
        assert extract(tosca) == {}

    def test_requirement_without_filter_ignored(self, extract):
        tosca = self._tosca(
            {"app": {"requirements": [{"host": {"node": "some-node"}}]}}
        )
        assert extract(tosca) == {}

    def test_node_with_filter_captured(self, extract):
        nf = {"$and": []}
        tosca = self._tosca({"app": {"requirements": [{"host": {"node_filter": nf}}]}})
        result = extract(tosca)
        assert "app" in result
        assert result["app"]["node_filter"] is nf

    def test_missing_service_template_returns_empty(self, extract):
        assert extract({}) == {}


# ---------------------------------------------------------------------------
# requirements.py — tosca_to_ask_dict
# ---------------------------------------------------------------------------


class TestToscaToAskDict:
    @pytest.fixture
    def to_ask(self):
        from sardou.requirements import tosca_to_ask_dict

        return tosca_to_ask_dict

    def test_no_node_filters_returns_empty(self, to_ask):
        tosca = {"service_template": {"node_templates": {"app": {"type": "X"}}}}
        assert to_ask(tosca) == {}

    def test_output_has_expression_and_colocated(self, to_ask):
        tosca = {
            "service_template": {
                "node_templates": {
                    "worker": {
                        "requirements": [
                            {
                                "host": {
                                    "node_filter": {
                                        "$and": [
                                            {
                                                "$greater_or_equal": [
                                                    {
                                                        "$get_property": [
                                                            "SELF",
                                                            "TARGET",
                                                            "CAPABILITY",
                                                            "host",
                                                            "num-cpus",
                                                        ]
                                                    },
                                                    4,
                                                ]
                                            }
                                        ]
                                    }
                                }
                            }
                        ]
                    }
                }
            }
        }
        result = to_ask(tosca)
        assert "worker" in result
        assert "expression" in result["worker"]
        assert "colocated" in result["worker"]
        assert "vals['host.num-cpus'] >= 4" in result["worker"]["expression"]

    def test_colocated_empty_when_no_policies(self, to_ask):
        tosca = {
            "service_template": {
                "node_templates": {
                    "svc": {"requirements": [{"host": {"node_filter": {"$and": []}}}]}
                }
            }
        }
        result = to_ask(tosca)
        assert result["svc"]["colocated"] == []

    def test_colocation_groups_representative_only(self, to_ask):
        tosca = {
            "service_template": {
                "node_templates": {
                    "a": {"requirements": [{"host": {"node_filter": {"$and": []}}}]},
                    "b": {"requirements": [{"host": {"node_filter": {"$and": []}}}]},
                },
                "policies": [
                    {
                        "col": {
                            "type": "swch:Scheduling.Colocation",
                            "targets": ["a", "b"],
                        }
                    }
                ],
            }
        }
        result = to_ask(tosca)
        assert "a" in result
        assert "b" not in result
        assert result["a"]["colocated"] == ["b"]

    def test_expression_is_evaluable_lambda(self, to_ask):
        tosca = {
            "service_template": {
                "node_templates": {
                    "svc": {
                        "requirements": [
                            {
                                "host": {
                                    "node_filter": {
                                        "$and": [
                                            {
                                                "$greater_or_equal": [
                                                    {
                                                        "$get_property": [
                                                            "SELF",
                                                            "TARGET",
                                                            "CAPABILITY",
                                                            "host",
                                                            "num-cpus",
                                                        ]
                                                    },
                                                    2,
                                                ]
                                            }
                                        ]
                                    }
                                }
                            }
                        ]
                    }
                }
            }
        }
        result = to_ask(tosca)
        fn = eval(result["svc"]["expression"])
        assert fn({"host.num-cpus": 4}) is True
        assert fn({"host.num-cpus": 1}) is False


# ---------------------------------------------------------------------------
# capacities.py — _unwrap
# ---------------------------------------------------------------------------


class TestUnwrap:
    @pytest.fixture
    def unwrap(self):
        from sardou.capacities import _unwrap

        return _unwrap

    def test_primitive_passthrough(self, unwrap):
        assert unwrap(42) == 42
        assert unwrap("hello") == "hello"

    def test_dollar_primitive(self, unwrap):
        assert unwrap({"$primitive": 7}) == 7

    def test_dollar_list(self, unwrap):
        assert unwrap({"$list": [{"$primitive": 1}, {"$primitive": 2}]}) == [1, 2]

    def test_plain_dict_passthrough(self, unwrap):
        assert unwrap({"a": 1, "b": 2}) == {"a": 1, "b": 2}

    def test_plain_dict_with_nested_primitive_not_recursed(self, unwrap):
        # _unwrap only unwraps at the top level; plain dicts are returned as-is
        result = unwrap({"key": {"$primitive": 99}})
        assert result == {"key": {"$primitive": 99}}


# ---------------------------------------------------------------------------
# capacities.py — _is_overall
# ---------------------------------------------------------------------------


class TestIsOverall:
    @pytest.fixture
    def is_overall(self):
        from sardou.capacities import _is_overall

        return _is_overall

    def test_matching_type_returns_true(self, is_overall):
        node = {"types": {"eu.swarmchestrate::OverallCapacity": {}}}
        assert is_overall(node) is True

    def test_non_matching_type_returns_false(self, is_overall):
        node = {"types": {"eu.swarmchestrate::Capacity": {}}}
        assert is_overall(node) is False

    def test_no_types_key_returns_false(self, is_overall):
        assert is_overall({}) is False


# ---------------------------------------------------------------------------
# capacities.py — extract_capacities
# ---------------------------------------------------------------------------


class TestExtractCapacities:
    @pytest.fixture
    def extract(self):
        from sardou.capacities import extract_capacities

        return extract_capacities

    def _node(self, types, caps=None):
        return {"types": types, "capabilities": caps or {}}

    def test_instance_based_returns_capacity_flavour(self, extract):
        nodes = {
            "small": self._node(
                {"eu.swch::Capacity": {}},
                {
                    "capacity": {"properties": {"instances": {"$primitive": 3}}},
                    "host": {"properties": {"num-cpus": {"$primitive": 2}}},
                },
            )
        }
        result = extract(nodes)
        assert "cloud_capacity_flavour" in result
        assert result["cloud_capacity_flavour"]["small"] == 3

    def test_overall_based_returns_capacity_raw(self, extract):
        nodes = {
            "pool": self._node(
                {"eu.swch::OverallCapacity": {}},
                {"capacity": {"properties": {"total": {"$primitive": 100}}}},
            )
        }
        result = extract(nodes)
        assert "cloud_capacity_raw" in result
        assert result["cloud_capacity_raw"]["total"] == 100

    def test_flavour_dict_populated(self, extract):
        nodes = {
            "medium": self._node(
                {"eu.swch::Capacity": {}},
                {"host": {"properties": {"mem-size": {"$primitive": "4"}}}},
            )
        }
        result = extract(nodes)
        assert "medium" in result["cloud_flavours"]
        assert result["cloud_flavours"]["medium"]["host"]["mem-size"] == "4"

    def test_empty_nodes_returns_empty_flavour(self, extract):
        result = extract({})
        assert "cloud_flavours" not in result

    def test_overall_node_excluded_from_flavour(self, extract):
        nodes = {
            "overall": self._node(
                {"eu.swch::OverallCapacity": {}},
                {"capacity": {"properties": {}}},
            )
        }
        result = extract(nodes)
        assert ("cloud_flavours" not in result) or (
            "overall" not in result["cloud_flavours"]
        )


# ---------------------------------------------------------------------------
# monitoring.py — extract_monitoring
# ---------------------------------------------------------------------------


class TestExtractMonitoring:
    @pytest.fixture
    def extract(self):
        from sardou.monitoring import extract_monitoring

        return extract_monitoring

    def test_node_with_metrics_and_slo(self, extract):
        nodes = {
            "app": {
                "capabilities": {
                    "metrics": {
                        "properties": {
                            "raw": [{"name": "cpu_util", "sensor": "Netdata"}],
                            "composite": [
                                {"name": "cpu_avg", "formula": "mean(cpu_util)"}
                            ],
                        }
                    },
                    "slo-constraints": {
                        "properties": {
                            "name": "cpu_slo",
                            "metric": "cpu_avg",
                            "operator": ">",
                            "threshold": 80.0,
                        }
                    },
                }
            }
        }
        result = extract(nodes)
        assert "app" in result
        assert "metrics" in result["app"]
        assert "slo-constraints" in result["app"]
        assert result["app"]["metrics"]["raw"][0]["name"] == "cpu_util"
        assert result["app"]["slo-constraints"]["threshold"] == 80.0

    def test_node_with_metrics_only(self, extract):
        nodes = {
            "svc": {
                "capabilities": {
                    "metrics": {"properties": {"raw": [{"name": "mem_usage"}]}}
                }
            }
        }
        result = extract(nodes)
        assert "svc" in result
        assert result["svc"]["metrics"]["raw"][0]["name"] == "mem_usage"
        assert result["svc"]["slo-constraints"] == {}

    def test_node_with_slo_only(self, extract):
        nodes = {
            "svc": {
                "capabilities": {
                    "slo-constraints": {
                        "properties": {
                            "name": "latency",
                            "operator": "<",
                            "threshold": 100,
                        }
                    }
                }
            }
        }
        result = extract(nodes)
        assert "svc" in result
        assert result["svc"]["slo-constraints"]["threshold"] == 100
        assert result["svc"]["metrics"] == {}

    def test_node_without_capabilities_excluded(self, extract):
        nodes = {"worker": {"properties": {"image": "nginx"}}}
        assert extract(nodes) == {}

    def test_node_with_unrelated_capabilities_excluded(self, extract):
        nodes = {"worker": {"capabilities": {"host": {"properties": {"num-cpus": 4}}}}}
        assert extract(nodes) == {}

    def test_multiple_nodes_mixed(self, extract):
        nodes = {
            "monitored": {
                "capabilities": {"metrics": {"properties": {"raw": [{"name": "cpu"}]}}}
            },
            "plain": {"properties": {"image": "nginx"}},
        }
        result = extract(nodes)
        assert "monitored" in result
        assert "plain" not in result

    def test_empty_nodes_returns_empty(self, extract):
        assert extract({}) == {}

    def test_get_monitoring_raises_on_non_sat(self):
        fake = object.__new__(SardouInternal)
        fake.kind = TemplateKind.CDT
        with pytest.raises(TypeError):
            fake.get_monitoring()


# ---------------------------------------------------------------------------
# Sardou public API (requires puccini-tosca)
# ---------------------------------------------------------------------------


@requires_puccini
@pytest.mark.parametrize("mode", ["path", "dict"])
class TestSardouCDTAPI:
    """Tests for the Sardou class used as a library, not via CLI."""

    @pytest.fixture
    def cloud_overall(self, mode):
        return _load_sardou(CDT_DIR / "cloud-overall.yaml", mode)

    @pytest.fixture
    def cloud_instances(self, mode):
        return _load_sardou(CDT_DIR / "cloud-instances.yaml", mode)

    def test_sardou_raises_on_missing_file(self, mode):
        if mode == "dict":
            pytest.skip("missing-file check only applies to path mode")
        with pytest.raises(FileNotFoundError):
            Sardou("/tmp/definitely_does_not_exist.yaml")

    def test_sardou_instance_has_path(self, cloud_overall, mode):
        if mode == "path":
            assert cloud_overall.path is not None
        else:
            assert cloud_overall.path is None

    def test_kind_cdt_for_capacity_template(self, cloud_overall, mode):
        assert cloud_overall.kind == TemplateKind.CDT

    def test_get_capacities_returns_dict(self, cloud_overall, mode):
        caps = cloud_overall.get_capacities()
        assert isinstance(caps, dict)
        assert "cloud_flavours" in caps

    def test_get_capacities_raises_on_sat(self, mode):
        """get_capacities() must raise TypeError when called on a SAT."""
        fake = object.__new__(SardouInternal)
        fake.kind = TemplateKind.SAT
        with pytest.raises(TypeError):
            fake.get_capacities()

    def test_get_monitoring_raises_on_cdt(self, mode):
        """get_monitoring() must raise TypeError when called on a CDT."""
        fake = object.__new__(SardouInternal)
        fake.kind = TemplateKind.CDT
        with pytest.raises(TypeError):
            fake.get_monitoring()

    def test_get_requirements_returns_dict(self, cloud_overall, mode):
        reqs = cloud_overall.get_requirements()
        assert isinstance(reqs, dict)

    def test_get_cluster_raises_type_error(self, cloud_overall, mode):
        with pytest.raises(TypeError):
            cloud_overall.get_cluster()

    def test_raw_attribute_is_accessible(self, cloud_overall, mode):
        assert hasattr(cloud_overall, "raw")

    def test_cloud_instances_parses(self, cloud_instances, mode):
        assert cloud_instances.kind == TemplateKind.CDT


@requires_puccini
@pytest.mark.parametrize("mode", ["path", "dict"])
class TestSardouSATAPI:
    """Tests for the Sardou class loaded with a SAT (service application template)."""

    @pytest.fixture
    def bookinfo(self, mode):
        return _load_sardou(SAT_DIR / "BookInfo.yaml", mode)

    def test_is_sat_true(self, bookinfo, mode):
        assert bookinfo.kind == TemplateKind.SAT

    def test_get_requirements_returns_dict(self, bookinfo, mode):
        reqs = bookinfo.get_requirements()
        assert isinstance(reqs, dict)

    def test_get_qos_returns_dict(self, bookinfo, mode):
        assert isinstance(bookinfo.get_qos(), dict)

    def test_get_monitoring_returns_dict(self, bookinfo, mode):
        monitoring = bookinfo.get_monitoring()
        assert isinstance(monitoring, dict)

    def test_get_capacities_raises_type_error(self, bookinfo, mode):
        with pytest.raises(TypeError):
            bookinfo.get_capacities()

    def test_get_cluster_raises_type_error(self, bookinfo, mode):
        with pytest.raises(TypeError):
            bookinfo.get_cluster()

    def test_raw_attribute_accessible(self, bookinfo, mode):
        assert hasattr(bookinfo, "raw")


# ---------------------------------------------------------------------------
# Integration
# ---------------------------------------------------------------------------


@requires_puccini
@pytest.mark.parametrize("mode", ["path", "dict"])
class TestSATTemplateOutput:
    """Verify that SAT templates produce meaningful, non-empty results."""

    @pytest.fixture
    def bookinfo(self, mode):
        return _load_sardou(SAT_DIR / "BookInfo.yaml", mode)

    def test_bookinfo_requirements_non_empty(self, bookinfo, mode):
        reqs = bookinfo.get_requirements()
        assert reqs, "BookInfo SAT should produce non-empty requirements"

    def test_bookinfo_requirements_have_capabilities(self, bookinfo, mode):
        reqs = bookinfo.get_requirements()
        for node_name, entry in reqs.items():
            assert "expression" in entry, (
                f"Requirement entry '{node_name}' should have expression"
            )
            assert entry["expression"], (
                f"Expression for '{node_name}' should be non-empty"
            )

    def test_bookinfo_requirements_have_metadata(self, bookinfo, mode):
        reqs = bookinfo.get_requirements()
        for node_name, entry in reqs.items():
            assert "colocated" in entry, (
                f"Requirement entry '{node_name}' should have colocated"
            )

    def test_bookinfo_colocation_applied(self, bookinfo):
        reqs = bookinfo.get_requirements()
        # details_v1 and productpage_v1 are colocated; only details_v1 should be representative
        has_colocated = any(len(e["colocated"]) > 0 for e in reqs.values())
        assert has_colocated, "At least one node should have colocated peers"

    def test_bookinfo_expressions_are_evaluable(self, bookinfo):
        reqs = bookinfo.get_requirements()
        for node_name, entry in reqs.items():
            fn = eval(entry["expression"])
            assert callable(fn), (
                f"Expression for '{node_name}' should be a callable lambda"
            )


@requires_puccini
@pytest.mark.parametrize("mode", ["path", "dict"])
class TestCDTTemplateOutput:
    """Verify that CDT templates produce meaningful, non-empty capacities."""

    @pytest.fixture(params=["cloud-overall.yaml", "cloud-instances.yaml"])
    def cdt(self, request, mode):
        return _load_sardou(CDT_DIR / request.param, mode)

    @pytest.fixture()
    def edge_cdt(self, mode):
        return _load_sardou(CDT_DIR / "edge.yaml", mode)

    def test_capacities_non_empty(self, cdt, mode):
        caps = cdt.get_capacities()
        assert caps, "CDT should produce non-empty capacities"

    def test_capacities_have_cloud_flavours(self, cdt, mode):
        caps = cdt.get_capacities()
        assert "cloud_flavours" in caps
        assert caps["cloud_flavours"], "CDT cloud_flavours dict should be non-empty"

    def test_edge_capacities_have_edge_instances(self, edge_cdt, mode):
        caps = edge_cdt.get_capacities()
        assert "edge_instances" in caps
        assert caps["edge_instances"], "CDT edge_instances dict should be non-empty"

    def test_flavour_entries_have_properties(self, cdt, mode):
        caps = cdt.get_capacities()
        for flavour_name, flavour in caps["cloud_flavours"].items():
            assert flavour, (
                f"Flavour '{flavour_name}' should have at least one capability"
            )

    def test_edge_flavour_entries_have_properties(self, edge_cdt, mode):
        caps = edge_cdt.get_capacities()
        for flavour_name, edge in caps["edge_instances"].items():
            assert edge, f"Edge '{flavour_name}' should have at least one capability"

    def test_cloud_flavour_capacity(self, cdt, mode):
        caps = cdt.get_capacities()
        assert caps.get("cloud_capacity_flavour") or caps.get("cloud_capacity_raw"), (
            "Cloud CDT should have a cloud_capacity_"
        )

    def test_edge_flavour_capacity_is_empty(self, edge_cdt, mode):
        caps = edge_cdt.get_capacities()
        assert not caps.get("cloud_capacity_flavour"), (
            "Edge CDT should not have cloud_capacity_flavour"
        )


OFFER_DIR = Path(__file__).parent / "templates" / "offer"

_rdt_scenarios = [
    ("cloud-instances", "cloud-instances.yaml", "offer-cloud.yaml"),
    ("cloud-overall", "cloud-overall.yaml", "offer-cloud.yaml"),
    ("edge", "edge.yaml", "offer-edge.yaml"),
]


@requires_puccini
@pytest.mark.parametrize(
    "cdt_file, offer_file",
    [(c, o) for _, c, o in _rdt_scenarios],
    ids=[name for name, _, _ in _rdt_scenarios],
)
class TestSardouRDTAPI:
    """Tests for RDT generation from a CDT and a selected offer."""

    @pytest.fixture
    def cdt(self, cdt_file):
        return Sardou(str(CDT_DIR / cdt_file))

    @pytest.fixture
    def offer(self, offer_file):
        with open(OFFER_DIR / offer_file) as f:
            return _yaml.load(f)

    @pytest.fixture
    def rdt(self, cdt, offer, tmp_path):
        out = str(tmp_path / "rdt.yaml")
        cdt.generate_rdt(offer, out)
        return Sardou(out)

    def test_rdt_has_one_node_per_offer_entry(self, rdt, offer):
        nodes = rdt.nodeTemplates._to_dict()
        assert set(nodes.keys()) == set(_get_offer_keys(offer))

    def test_rdt_nodes_keyed_by_offer_key(self, rdt, offer):
        nodes = rdt.nodeTemplates._to_dict()
        for key in _get_offer_keys(offer):
            assert key in nodes

    def test_rdt_node_has_count(self, rdt):
        nodes = rdt.nodeTemplates._to_dict()
        for node in nodes.values():
            assert "count" in node

    def test_duplicate_res_id_produces_separate_nodes(self, rdt, offer):
        """Nodes with same res_id but different keys should be separate."""
        nodes = rdt.nodeTemplates._to_dict()
        keys = _get_offer_keys(offer)
        for key in keys:
            assert key in nodes

    def test_rdt_node_inherits_cdt_type(self, cdt, rdt, offer):
        cdt_nodes = cdt.nodeTemplates._to_dict()
        rdt_nodes = rdt.nodeTemplates._to_dict()
        for key in _get_offer_keys(offer):
            res_id = _get_offer_res_id(offer, key)
            cdt_types = set(cdt_nodes[res_id]["types"])
            rdt_types = set(rdt_nodes[key]["types"])
            assert cdt_types.issubset(rdt_types)

    def test_rdt_metadata_has_kind(self, rdt):
        assert rdt.metadata._to_dict().get("kind") == "RDT"

    def test_invalid_res_id_raises(self, cdt, offer_file):  # noqa: ARG002
        # Simulate new-offer.json structure with invalid res_id
        bad_offer = {
            "details_v1": {
                "bad-resource": {"ids": {"res_id": "nonexistent"}, "count": 1}
            }
        }
        with pytest.raises(KeyError):
            cdt.generate_rdt(bad_offer, "/dev/null")

    def test_offer_properties_merged_into_node(self, cdt, offer_file, tmp_path):  # noqa: ARG002
        """Properties from offer are added to the RDT node."""
        res_id = next(iter(cdt.nodeTemplates._to_dict()))
        offer = {
            "ms1": {
                "offer-key": {
                    "ids": {"res_id": res_id},
                    "count": 1,
                    "properties": {
                        "ingress-rules": [{"from": 80, "to": 80, "protocol": "TCP"}]
                    },
                }
            }
        }
        out = str(tmp_path / "rdt.yaml")
        cdt.generate_rdt(offer, out)
        rdt = Sardou(out)
        props = rdt.nodeTemplates._to_dict()["offer-key"]["properties"]
        assert "ingress-rules" in props

    def test_offer_properties_do_not_overwrite_cdt_properties(
        self, cdt, offer_file, tmp_path
    ):  # noqa: ARG002
        """Existing CDT node properties are not overwritten by offer properties."""
        cdt_nodes = cdt.nodeTemplates._to_dict()
        res_id = next(k for k, v in cdt_nodes.items() if v.get("properties"))
        existing_prop = next(iter(cdt_nodes[res_id]["properties"]))
        offer = {
            "ms1": {
                "offer-key": {
                    "ids": {"res_id": res_id},
                    "count": 1,
                    "properties": {existing_prop: "OVERWRITTEN"},
                }
            }
        }
        out = str(tmp_path / "rdt.yaml")
        cdt.generate_rdt(offer, out)
        rdt = Sardou(out)
        props = rdt.nodeTemplates._to_dict()["offer-key"]["properties"]
        assert props[existing_prop] != "OVERWRITTEN"


# Helper functions for offer key extraction
def _get_offer_keys(offer):
    if isinstance(offer, list):
        offer = offer[0]
    if all(isinstance(v, dict) and "res_id" in v for v in offer.values()):
        return list(offer.keys())
    keys = []
    for ms_data in offer.values():
        if isinstance(ms_data, dict):
            for offer_key, offer_data in ms_data.items():
                if isinstance(offer_data, dict) and "ids" in offer_data:
                    keys.append(offer_key)
    return keys


def _get_offer_res_id(offer, key):
    if isinstance(offer, list):
        offer = offer[0]
    if key in offer and "res_id" in offer[key]:
        return offer[key]["res_id"]
    for ms_data in offer.values():
        if isinstance(ms_data, dict):
            if key in ms_data and "ids" in ms_data[key]:
                return ms_data[key]["ids"].get("res_id")
    return None



import copy

from ruamel.yaml import YAML

rdt_yaml = YAML()
rdt_yaml.default_flow_style = False


def _validate_offer_against_cdt(selected_offer: dict, cdt_nodes: dict) -> None:
    for ms_name, ms_data in selected_offer.items():
        if not isinstance(ms_data, dict):
            continue
        for offer_key, offer_data in ms_data.items():
            if not isinstance(offer_data, dict):
                continue
            ids = offer_data.get("ids", {})
            res_id = ids.get("res_id")
            if not res_id:
                continue

            if res_id not in cdt_nodes:
                raise KeyError(
                    f"CDT validation failed: res_id '{res_id}' from offer '{offer_key}' "
                    f"does not match any node in the CDT. "
                    f"The CDT may have been modified since this offer was generated."
                )


def generate_rdt(template, selected_offer: dict, output_path: str = "rdt.yaml") -> dict:
    source = template.raw._to_dict()

    rdt = {}
    rdt["tosca_definitions_version"] = source.get(
        "tosca_definitions_version", "tosca_2_0"
    )
    rdt["description"] = source.get("description", "Resource Definition Template")
    rdt["metadata"] = copy.deepcopy(source.get("metadata", {}))
    rdt["metadata"]["kind"] = "RDT"
    rdt["imports"] = copy.deepcopy(source.get("imports", []))

    try:
        cdt_nodes = source.get("service_template", {})["node_templates"]
    except KeyError:
        raise ValueError("Invalid CDT: 'node_templates' not found")

    _validate_offer_against_cdt(selected_offer, cdt_nodes)

    cdt_node_types = source.get("node_types", {})
    new_node_templates = {}
    used_local_types = set()

    for ms_name, ms_data in selected_offer.items():
        if not isinstance(ms_data, dict):
            continue

        for offer_key, offer_data in ms_data.items():
            if not isinstance(offer_data, dict):
                continue

            ids = offer_data.get("ids", {})
            res_id = ids.get("res_id")
            if not res_id:
                continue

            node_key = res_id

            node_type = cdt_nodes[node_key]["type"]
            node = copy.deepcopy(cdt_nodes[node_key])

            ms_id = ids.get("ms_id", "")
            node.setdefault("metadata", {})["ms_id"] = ms_id

            offer_props = offer_data.get("properties", {})
            if offer_props:
                node.setdefault("properties", {}).update(offer_props)

            node["count"] = offer_data.get("count", 1)

            new_node_templates[offer_key] = node

            if node_type in cdt_node_types:
                used_local_types.add(node_type)

    if used_local_types:
        rdt["node_types"] = {
            type_name: copy.deepcopy(cdt_node_types[type_name])
            for type_name in used_local_types
        }

    rdt["service_template"] = {"node_templates": new_node_templates}

    with open(output_path, "w") as f:
        rdt_yaml.width = 4096
        for key, value in rdt.items():
            rdt_yaml.dump({key: value}, f)
            f.write("\n")

    return rdt



import logging
import subprocess
from enum import Enum
from functools import wraps
from pathlib import Path
from tempfile import NamedTemporaryFile

from ruamel.yaml import YAML

logger = logging.getLogger(__name__)


class TemplateKind(Enum):
    SAT = "sat"
    CDT = "cdt"
    RDT = "rdt"
    TDT = "tdt"


def requires_kind(kind):
    def decorator(fn):
        @wraps(fn)
        def wrapper(self, *args, **kwargs):
            if self.kind != kind:
                raise TypeError(f"{fn.__name__}() requires a {kind.value.upper()}")
            return fn(self, *args, **kwargs)

        return wrapper

    return decorator


PUCCINI_CMD = "/usr/bin/puccini-tosca"
PUCCINI_FLAGS = ["-x", "data_types.string.permissive"]

# Read and update YAML using ruamel.yaml
yaml = YAML()


def prevalidate(input_data):
    if isinstance(input_data, Path):
        if not input_data.exists():
            logger.error(f"File does not exist: {input_data}")
            return False
        try:
            with input_data.open("r") as f:
                data = yaml.load(f)
        except Exception as e:
            logger.error(f"Error reading YAML file {input_data}: {e}")
            return False
    elif isinstance(input_data, dict):
        data = input_data
    else:
        try:
            data = yaml.load(input_data)
        except Exception as e:
            logger.error(f"Error parsing YAML content: {e}")
            return False

    if not data:
        logger.error("No YAML content found")
        return False

    imports = data.get("imports", [])
    template = data.get("service_template", {})

    for imp in imports:
        if isinstance(imp, dict) and "profile" in imp:
            imp["url"] = imp.pop("profile")

    for _, node in template.get("node_templates", {}).items():
        node.pop("node_filter", None)

    return data


def classify_template(template) -> TemplateKind:
    """Classify a parsed template as SAT, CDT, RDT, or TDT."""

    kind_str = template._to_dict()["metadata"].get("kind", "").lower()
    try:
        return TemplateKind(kind_str)
    except ValueError:
        pass

    nodes = template.nodeTemplates._to_dict()
    if not nodes:
        return TemplateKind.TDT

    has_capacity = any(
        any(k.endswith("::Capacity") for k in node.get("types", {}))
        for node in nodes.values()
    )
    has_microservice = any(
        any(k.endswith("::Microservice") for k in node.get("types", {}))
        for node in nodes.values()
    )

    if has_capacity and has_microservice:
        raise ValueError(
            "Invalid template: cannot have both Capacity and Microservice definitions"
        )

    if has_capacity:
        return TemplateKind.CDT

    return TemplateKind.SAT


def validate_template(input_data) -> bool:
    # will run the puccini-tosca parse <with flag>
    yaml_data = prevalidate(input_data)

    if isinstance(input_data, Path):
        file_label = str(input_data)
    elif isinstance(input_data, dict):
        file_label = "(dict content)"
    else:
        file_label = "(string content)"

    # open a temp file
    with NamedTemporaryFile() as temp_file:
        yaml.dump(yaml_data, temp_file)

        try:
            result = subprocess.run(
                [PUCCINI_CMD, "parse", str(temp_file.name)] + PUCCINI_FLAGS,
                capture_output=True,
                text=True,
            )
            if result.returncode == 0:
                logger.info(f"Processed successfully: {file_label}")
                return result
            else:
                logger.error(f"Failed to process: {file_label}")
                logger.error(result.stderr.strip() or result.stdout.strip())
                return None

        except FileNotFoundError:
            raise FileNotFoundError(
                f"Puccini not found at {PUCCINI_CMD}. Please install it first."
            )



from typing import Optional

from ruamel.yaml import YAML

from sardou import Sardou

# Read and update YAML using ruamel.yaml
yaml = YAML()


def get_kubernetes_manifest(
    tosca_yaml: str, image_pull_secret: Optional[str] = None
) -> list:
    """
    Convert a TOSCA template string into Kubernetes manifests (Deployment + Service).
    Always injects external imagePullSecret. Handles:
    - replicas
    - env
    - args
    - ports
    - volumes
    - nodeSelector
    - imagePullSecrets
    """

    # Load and validate TOSCA

    tosca = Sardou(tosca_yaml)
    node_templates = (
        tosca.raw._to_dict().get("service_template", {}).get("node_templates", {})
    )

    if not node_templates:
        raise ValueError("No node_templates found in TOSCA YAML")

    manifests = []
    # Iterate over nodes
    for name, node in node_templates.items():
        try:
            # Sanitize name for Kubernetes RFC 1123 compliance (replace underscores with hyphens)
            k3s_name = name.replace("_", "-")

            node_type = node.get("type", "")
            if not node_type.endswith("Microservice"):
                continue

            props = node.get("properties", {}) or {}
            requirements = node.get("requirements", []) or []

            # Basic fields
            image = props.get("image")
            if not image:
                raise ValueError(f"{name}: missing image")

            replicas = int(props.get("replicas", 1))
            args = props.get("args", []) or []

            # Environment variables
            env_list = []
            for e in props.get("env", []) or []:
                if isinstance(e, dict) and "name" in e:
                    env_list.append(
                        {"name": str(e["name"]), "value": str(e.get("value", ""))}
                    )

            # PORTS
            container_ports = []
            service_ports = []

            for p in props.get("ports", []) or []:
                port = int(p.get("port", p.get("targetPort")))
                target = int(p.get("targetPort", port))
                protocol = str(p.get("protocol", "TCP")).upper()
                node_port = p.get("nodePort")

                # Deployment container ports
                container_ports.append({"containerPort": target, "protocol": protocol})

                # Service ports
                sp = {
                    "name": p.get("name", f"port-{port}"),
                    "port": port,
                    "targetPort": target,
                    "protocol": protocol,
                }

                if node_port:
                    sp["nodePort"] = int(node_port)

                service_ports.append(sp)

            # Volumes
            volumes = []
            volume_mounts = []

            for r in requirements:
                if "volume" in r:
                    vol_name = r["volume"]
                    vol_def = node_templates.get(vol_name, {})
                    props_v = vol_def.get("properties", {})
                    path = props_v.get("path")

                    if path:
                        volumes.append(
                            {
                                "name": vol_name,
                                "hostPath": {"path": path, "type": "DirectoryOrCreate"},
                            }
                        )
                        volume_mounts.append({"name": vol_name, "mountPath": path})

            # Node labels / nodeSelector
            labels = {}

            for r in requirements:
                if "host" in r:
                    node_filter = r["host"].get("node_filter", {})

                    for cond in node_filter.get("$and", []):
                        if "$equal" in cond:
                            left, right = cond["$equal"]
                            if isinstance(left, dict) and "$get_property" in left:
                                path = left["$get_property"]
                                if (
                                    isinstance(path, list)
                                    and len(path) >= 6
                                    and path[:5]
                                    == [
                                        "SELF",
                                        "TARGET",
                                        "CAPABILITY",
                                        "resource",
                                        "labels",
                                    ]
                                ):
                                    key = path[5]
                                    value = right
                                    if value not in (None, "", []):
                                        labels[key] = value
            # Deployment (standard field order)
            deployment = {
                "apiVersion": "apps/v1",
                "kind": "Deployment",
                "metadata": {"name": k3s_name},
                "spec": {
                    "replicas": replicas,
                    "selector": {"matchLabels": {"app": k3s_name}},
                    "template": {
                        "metadata": {"labels": {"app": k3s_name}},
                        "spec": {
                            **(
                                {"imagePullSecrets": [{"name": image_pull_secret}]}
                                if image_pull_secret
                                else {}
                            ),
                            **({"nodeSelector": labels} if labels else {}),
                            "containers": [
                                {
                                    "name": k3s_name,
                                    "image": image,
                                    **({"args": args} if args else {}),
                                    **({"env": env_list} if env_list else {}),
                                    **(
                                        {"ports": container_ports}
                                        if container_ports
                                        else {}
                                    ),
                                    **(
                                        {"volumeMounts": volume_mounts}
                                        if volume_mounts
                                        else {}
                                    ),
                                }
                            ],
                            **({"volumes": volumes} if volumes else {}),
                        },
                    },
                },
            }

            manifests.append(deployment)

            # Service (only if ports are defined)
            if service_ports:
                svc_type = (
                    "NodePort"
                    if any("nodePort" in p for p in service_ports)
                    else "ClusterIP"
                )

                service = {
                    "apiVersion": "v1",
                    "kind": "Service",
                    "metadata": {"name": k3s_name},
                    "spec": {
                        "type": svc_type,
                        "selector": {"app": k3s_name},
                        "ports": service_ports,
                    },
                }

                manifests.append(service)
        except Exception as e:
            print(f"Skipping node '{name}' due to error: {e}")
            continue

    return manifests



import sys
from pathlib import Path

from sardou.validation import validate_template


def main():
    templates_dir = Path("templates")
    yaml_files = list(
        templates_dir.rglob("*.yaml")
    )  # finds all YAML files inside subfolders

    if not yaml_files:
        print("No YAML files found under templates/")
        sys.exit(0)

    success, fail = 0, 0
    for file in yaml_files:
        print(f"\nProcessing {file}...\n")
        if validate_template(file):
            success += 1
        else:
            fail += 1

    print("============================")
    print(f"{success} Successful")
    print(f"{fail} Failed")
    print("============================")

    if fail > 0:
        sys.exit(1)


if __name__ == "__main__":
    main()
